from notist._notifiers.base import BaseNotifier
from notist._notifiers.discord import DiscordNotifier
from notist._notifiers.slack import SlackNotifier

__all__ = [
    "BaseNotifier",
    "DiscordNotifier",
    "SlackNotifier",
]



from __future__ import annotations

import functools
import textwrap
from datetime import timedelta
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from collections.abc import Callable

# NOTE: Python 3.12+ (PEP 695) supports inline type parameter syntax.
# After dropping Python 3.11 support, update this to use that instead.
# See:
#   - https://peps.python.org/pep-0695/
#   - https://docs.python.org/3/reference/compound_stmts.html#type-params
T = TypeVar("T", bound=type[Any])


def _clone_function(fn: Callable[..., Any]) -> Callable[..., Any]:
    """Clone a function to avoid modifying the original."""

    @functools.wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return fn(*args, **kwargs)

    return wrapper


def extend_method_docstring(
    additions: dict[str, str | Callable[[T], str]],
) -> Callable[[T], T]:
    """
    Class decorator factory that appends extra text to inherited methods' docstrings.
    `additions` should map method names to the snippet you want appended.
    """

    def decorator(cls: T) -> T:
        for name, doc in additions.items():
            if not hasattr(cls, name):
                continue
            base_cls = next((b for b in cls.__mro__[1:] if hasattr(b, name)), cls)
            method = getattr(base_cls if name in cls.__dict__ else cls, name)
            base = method.__doc__ or ""
            doc_str = doc(cls) if callable(doc) else doc
            extra = textwrap.dedent(doc_str).strip()
            new_doc = base + "\n\n" + textwrap.indent(extra, " " * 8)
            if name in cls.__dict__:
                cls.__dict__[name].__doc__ = new_doc
            else:
                new_method = _clone_function(method)
                new_method.__doc__ = new_doc
                setattr(cls, name, new_method)
        return cls

    return decorator


def format_timedelta(td: timedelta) -> str:
    days = td.days
    hours, remainder = divmod(td.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    parts = [
        days and f"{days}d",
        hours and f"{hours}h",
        minutes and f"{minutes}m",
        seconds and f"{seconds}s",
    ]
    return " ".join([p for p in parts if p]) or "0s"



from __future__ import annotations

from datetime import timedelta

import pytest

from notist._utils import format_timedelta


@pytest.mark.parametrize(
    "delta, expected",
    [
        (timedelta(seconds=3), "3s"),
        (timedelta(minutes=2, seconds=3), "2m 3s"),
        (timedelta(hours=1), "1h"),
        (timedelta(days=2, hours=1), "2d 1h"),
        (timedelta(), "0s"),
    ],
)
def test_format_timedelta(delta: timedelta, expected: str) -> None:
    assert format_timedelta(delta) == expected



from __future__ import annotations

import functools
import inspect
import linecache
import traceback
from collections.abc import Callable, Iterator
from contextlib import AbstractContextManager, ContextDecorator
from datetime import datetime
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

from notist import _log
from notist._log import (
    LEVEL_ORDER,
    RESET,
    LevelStr,
    fg256,
)
from notist._log import Glyph as _G
from notist._log import SpecialToken as _S
from notist._utils import format_timedelta

if TYPE_CHECKING:
    import sys
    from collections.abc import Generator, Iterable
    from types import TracebackType

    from notist._notifiers.base import _SendFnPartial

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self


# NOTE: Python 3.12+ (PEP 695) supports inline type parameter syntax.
# After dropping Python 3.11 support, update this to use that instead.
# See:
#   - https://peps.python.org/pep-0695/
#   - https://docs.python.org/3/reference/compound_stmts.html#type-params
T = TypeVar("T")
_F = TypeVar("_F", bound=Callable[..., Any])


class Watch(ContextDecorator, AbstractContextManager):
    def __init__(
        self,
        send_fn: _SendFnPartial,
        params: str | list[str] | None = None,
        label: str | None = None,
        callsite_level: LevelStr = "error",
        callsite_context_before: int = 1,
        callsite_context_after: int = 4,
        combined: int = 0,
    ) -> None:
        self._send = send_fn
        self._params = [params] if isinstance(params, str) else params or []
        self._param_vals: dict[str, Any] | None = None
        self._start: datetime | None = None
        self._label = label
        self._callsite_level = callsite_level
        self._callsite_context_before = callsite_context_before
        self._callsite_context_after = callsite_context_after
        self._combined: int = combined
        self._target: str | None = None
        self._called_from: str | None = None
        self._defined_at: str | None = None
        self._is_fn = False
        self._filename: str | None = None
        self._lineno: int | None = None

    def __enter__(self) -> Self:
        self._start = datetime.now()
        if not self._is_fn and self._params and self._send.config.verbose:
            _log.warn(
                "Parameters can only be captured when used as a decorator on a function. Ignoring 'params' argument."
            )

        f = (f0 := inspect.currentframe()) and f0.f_back
        if self._is_fn:
            for _ in range(max(1, self._combined) * 2):
                f = f and f.f_back
        elif self._combined:
            f = f and f.f_back

        self._filename = f and f.f_code.co_filename
        fnname = f and f.f_code.co_name
        self._lineno = f and f.f_lineno
        module = f and f.f_globals.get("__name__", "<unknown>")

        module_fname = (
            f"{_S.BT_ALW}{module}.{fnname}{_S.BT_ALW}"
            if fnname != "<module>"
            else f"{_S.BT_ALW}{module}{_S.BT_ALW}"
        )
        if self._is_fn:
            self._called_from = f"{module_fname} @ {self._filename}:{self._lineno}"
        else:
            self._called_from = f"{self._filename}:{self._lineno}"
            self._target = f"code block in {module_fname}"

        message = f"Start watching{self._details()}"
        self._send(message)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        assert self._start
        end = datetime.now()
        et_msg_raw = f"Execution time: {format_timedelta(end - self._start)}"
        et_msg = fg256(8) + " " + _G.CBULLET + " " + et_msg_raw + RESET
        exc_only = "".join(traceback.format_exception_only(exc_type, exc_val)).strip()
        if exc_type:
            tb_str = "".join(traceback.format_exception(exc_type, exc_val, tb))
            error_msg = (
                f"Error while watching{self._details('error', exc_only)}\n{et_msg}"
            )
            self._send(error_msg, tb=tb_str, level="error")
        else:
            msg = f"End watching{self._details()}\n{et_msg}"
            self._send(msg)

    def _details(self, level: LevelStr = "info", message: str | None = None) -> str:
        assert self._called_from is not None
        target = (
            f" {fg256(45)}<{self._target}> [label: {self._label}]{RESET}"
            if self._label
            else f" {fg256(45)}<{self._target}>{RESET}"
        )
        called_lines = (
            (LEVEL_ORDER[self._callsite_level] <= LEVEL_ORDER[level])
            and _get_called_lines_str(
                self._filename,
                self._lineno,
                self._callsite_context_before,
                self._callsite_context_after,
                message,
            )
            or None
        )
        if self._is_fn:
            assert self._defined_at is not None
            defined_at = f" {fg256(8)}{_G.RARROWF} Defined at: {fg256(12)}{self._defined_at}{RESET}"
            called_from = f" {fg256(8)}{_G.RARROWF} Called from: {fg256(12)}{self._called_from}{RESET}"
            params = (self._param_vals or None) and (
                f"   {fg256(8)}{_G.RARROWF}{_G.RARROWF} With params: "
                + ", ".join(f"{k}={v!r}" for k, v in self._param_vals.items())
            )
            return "\n".join(
                filter(None, [target, defined_at, called_from, params, called_lines])
            )
        else:
            called_from = (
                f" {fg256(8)}{_G.RARROWF} at: {fg256(12)}{self._called_from}{RESET}"
            )
            return "\n".join(filter(None, [target, called_from, called_lines]))

    def __call__(self, fn: _F) -> _F:
        self._is_fn = True

        orig_fn = fn
        while hasattr(orig_fn, "__wrapped__"):
            orig_fn = orig_fn.__wrapped__
        filename = orig_fn.__code__.co_filename
        lineno = orig_fn.__code__.co_firstlineno
        module = orig_fn.__module__
        qualname = orig_fn.__qualname__
        self._target = f"function {_S.BT_ALW}{module}.{qualname}{_S.BT_ALW}"
        self._defined_at = f"{filename}:{lineno}"

        @functools.wraps(fn)
        def _wrapped(*args: Any, **kwargs: Any) -> Any:
            bound = inspect.signature(fn).bind(*args, **kwargs)
            bound.apply_defaults()
            missing = [p for p in self._params if p not in bound.arguments]
            if missing and self._send.config.verbose:
                _log.warn(
                    f"Parameters {missing} not found in function arguments. Skipping capturing their values."
                )
            self._param_vals = {
                p: bound.arguments[p] for p in self._params if p in bound.arguments
            }
            return super(Watch, self).__call__(fn)(*args, **kwargs)

        return cast(_F, _wrapped)


class IterableWatch(AbstractContextManager, Generic[T]):
    def __init__(
        self,
        iterable: Iterable[T],
        step: int,
        total: int | None,
        send_fn: Callable[..., None],
        label: str | None,
        callsite_level: LevelStr = "error",
        callsite_context_before: int = 1,
        callsite_context_after: int = 4,
        class_name: str | None = None,
        object_id: int | None = None,
    ) -> None:
        self._iterable = iterable
        self._step = step
        self._total = (
            len_fn() if (len_fn := getattr(iterable, "__len__", None)) else total
        )
        self._send = send_fn
        self._label = label
        self._callsite_level = callsite_level
        self._callsite_context_before = callsite_context_before
        self._callsite_context_after = callsite_context_after
        self._iterable_object_str = f"{class_name or iterable.__class__.__name__} object at {object_id or hex(id(iterable))}"
        self._start: datetime | None = datetime.now()
        self._count: int | None = None
        self._prev_start: datetime | None = None
        self._cur_range_start: int | None = None
        self._cur_range_end: int | None = None
        self._filename: str | None = None
        self._lineno: int | None = None
        self._called_from: str | None = None

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        assert self._count is not None
        if not exc_type:
            return
        tb_str = "".join(traceback.format_exception(exc_type, exc_val, tb))
        exc_only = "".join(traceback.format_exception_only(exc_type, exc_val)).strip()
        message = (
            "Error while processing "
            + self._item_message
            + (f"of {self._total} " if self._total is not None else "")
            + "from"
            + self._details("error", exc_only)
            + "\n"
            + self._et_message
        )
        self._send(message, tb=tb_str, level="error")

    def __iter__(self) -> Iterator[T]:
        return self._gen()

    def _gen(self) -> Generator[T, None, None]:
        self._on_iter_start()
        for item in self._iterable:
            self._on_step_start()
            yield item
            self._on_step_end()
        self._on_iter_end()

    def _send_end_message(self) -> None:
        self._send(
            "Processed "
            + self._item_message
            + (f"of {self._total} " if self._total is not None else "")
            + f"from{self._details()}\n"
            + self._et_message
        )

    def _on_iter_start(self) -> None:
        self._set_callsite_info()
        self._start = datetime.now()
        self._send(f"Start watching{self._details()}")

    def _on_iter_end(self) -> None:
        assert self._count is not None
        if (self._count + 1) % self._step:
            self._cur_range_end = self._count + 1
            self._send_end_message()

        assert self._start is not None
        end = datetime.now()
        self._send(
            f"End watching{self._details()}\n"
            f"{fg256(8)} {_G.CBULLET} "
            f"Total execution time: {format_timedelta(end - self._start)}"
        )

    def _on_step_start(self) -> None:
        self._count = 0 if self._count is None else self._count + 1
        assert self._count is not None
        if self._count % self._step:
            return
        self._cur_range_start = self._count + 1
        self._cur_range_end = min(self._count + self._step, self._total or 1 << 30)
        self._prev_start = datetime.now()
        self._send(
            "Processing "
            + self._item_message
            + (f"of {self._total} " if self._total is not None else "")
            + f"from{self._details()}"
        )

    def _on_step_end(self) -> None:
        if self._count is None or (self._count + 1) % self._step:
            return
        self._send_end_message()

    def _set_callsite_info(self) -> None:
        f = (f0 := inspect.currentframe()) and (f1 := f0.f_back) and f1.f_back
        f = f and f.f_back
        self._filename = f and f.f_code.co_filename
        fnname = f and f.f_code.co_name
        self._lineno = f and f.f_lineno
        module = f and f.f_globals.get("__name__", "<unknown>")
        module_fname = f"{module}.{fnname}" if fnname != "<module>" else f"{module}"
        self._called_from = (
            f"{_S.BT_ALW}{module_fname}{_S.BT_ALW} @ {self._filename}:{self._lineno}"
        )

    def _details(self, level: LevelStr = "info", message: str | None = None) -> str:
        target = (
            f" {fg256(45)}{_S.BT_MSG}<{self._iterable_object_str}>{_S.BT_MSG} [label: {self._label}]{RESET}"
            if self._label
            else f" {fg256(45)}{_S.BT_MSG}<{self._iterable_object_str}>{_S.BT_MSG}{RESET}"
        )
        called_from = (
            f" {fg256(8)}{_G.RARROWF} in: {fg256(12)}{self._called_from}{RESET}"
        )
        called_lines = (
            (LEVEL_ORDER[self._callsite_level] <= LEVEL_ORDER[level])
            and _get_called_lines_str(
                self._filename,
                self._lineno,
                self._callsite_context_before,
                self._callsite_context_after,
                message,
            )
            or None
        )
        return "\n".join(filter(None, [target, called_from, called_lines]))

    @property
    def _et_message(self) -> str:
        end = datetime.now()
        assert self._prev_start is not None
        assert self._start is not None
        return (
            f"{fg256(8)} {_G.CBULLET} Execution time for "
            + self._item_message.rstrip()
            + (f" of {self._total}" if self._total is not None else "")
            + f": {format_timedelta(end - self._prev_start)}{RESET}\n"
            + f"{fg256(8)} {_G.CBULLET}"
            + f" Total execution time: {format_timedelta(end - self._start)}{RESET}"
        )

    @property
    def _item_message(self) -> str:
        assert self._cur_range_start is not None
        assert self._cur_range_end is not None
        return (
            f"item {self._cur_range_start} "
            if self._step == 1
            else f"items {self._cur_range_start}–{self._cur_range_end} "
        )


def _get_called_lines_str(
    filename: str | None,
    lineno: int | None,
    callsite_context_before: int,
    callsite_context_after: int,
    message: str | None = None,
) -> str | None:
    if not filename or lineno is None:
        return None
    called_lines = [
        (num := lineno + i, linecache.getline(filename, num).rstrip())
        for i in range(-callsite_context_before, callsite_context_after)
    ]
    w = len(str(called_lines[-1][0]))
    called_lines_ls = [
        f"  {fg256(20)}{i:>{w}d} {fg256(57)}{_G.V}{RESET} {line}"
        for i, line in called_lines
    ]
    wnum = len(line := called_lines[callsite_context_before][1]) - (
        snum := len(line.lstrip())
    )
    underline = (
        fg256(45) + _G.TL + _G.H * (3 + wnum) + _G.TDH * 2 + " " + _G.H * snum + RESET
        if message
        else fg256(45) + " " * 7 + " " * wnum + _G.H * snum + RESET
    )
    return (
        _S.FNO_MSG
        + "\n".join(
            called_lines_ls[: (idx := callsite_context_before + 1)]
            + [underline]
            + [
                f"{fg256(45)}{_G.V}{RESET}{line[1:]}" if message else line
                for line in called_lines_ls[idx:]
            ]
            + (
                [
                    f"{fg256(45)}{_G.BL}{_G.H}{_G.RARROWP}{RESET} {fg256(197)}{message}{RESET}"
                ]
                if message
                else []
            )
        )
        + _S.FNC_MSG
    )



from __future__ import annotations

import os
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Protocol, TypedDict, TypeVar, overload

from notist import _log
from notist._log import LevelStr, prepare_for_message
from notist._watch import IterableWatch, Watch

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from types import ModuleType, TracebackType

    if sys.version_info >= (3, 12):
        from typing import Unpack
    else:
        from typing_extensions import Unpack


# NOTE: Python 3.12+ (PEP 695) supports inline type parameter syntax.
# After dropping Python 3.11 support, update this to use that instead.
# See:
#   - https://peps.python.org/pep-0695/
#   - https://docs.python.org/3/reference/compound_stmts.html#type-params
P = ParamSpec("P")
R = TypeVar("R")
T_co = TypeVar("T_co", covariant=True)
T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Any])


# This protocol guarantees to static checkers (e.g. mypy) that any implementing
# object have  `__enter__`, `__exit__` and `__call__`.
# Otherwise, users applying these contexts would get mypy errors because the type
# system wouldn't know these methods exist.
class ContextManagerDecorator(Protocol[F]):
    """Protocol for objects that can be used as context managers and decorators."""

    def __enter__(self) -> Self: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        tb: TracebackType | None,
        /,
    ) -> None: ...
    def __call__(self, fn: F, /) -> F: ...


class ContextManagerIterator(Protocol[T_co]):
    """Protocol for objects that can be used as context managers and iterators."""

    def __enter__(self) -> Self: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        tb: TracebackType | None,
        /,
    ) -> None: ...
    def __iter__(self) -> Iterator[T_co]: ...


@dataclass(frozen=True)
class _SendConfig:
    channel: str | None = None
    mention_to: str | None = None
    mention_level: LevelStr = "error"
    mention_if_ends: bool = True
    verbose: bool = True
    disable: bool = False


@dataclass(frozen=True)
class _SendFnPartial:
    fn: Callable[[_SendConfig, str, str | None, LevelStr, str], None]
    config: _SendConfig

    def __call__(
        self,
        message: str,
        tb: str | None = None,
        level: LevelStr = "info",
        prefix: str = "",
    ) -> None:
        self.fn(self.config, message, tb, level, prefix)


class SendOptions(TypedDict, total=False):
    """
    Additional options for :meth:`~BaseNotifier.watch` and :meth:`~BaseNotifier.register`.
    """

    #: Optional label for the message. Included in notification messages and log entries.
    label: str | None
    #: Override the default channel for notifications.
    channel: str | None
    #: Override the default user to mention on notification.
    mention_to: str | None
    #: Override the default mention threshold level.
    mention_level: LevelStr | None
    #: Override the default setting for whether to mention at the end of the watch.
    mention_if_ends: bool | None
    #: Override the default call-site source snippet threshold level.
    callsite_level: LevelStr | None
    #: Number of lines of context to include before the call site. Default is 1.
    callsite_context_before: int
    #: Number of lines of context to include after the call site. Default is 4.
    callsite_context_after: int
    #: Override the default verbosity setting.
    verbose: bool | None
    #: Override the default disable flag.
    disable: bool | None


@dataclass(frozen=True)
class _SendOptions:
    """
    Internal version of SendOptions with default values filled in.
    """

    label: str | None = None
    channel: str | None = None
    mention_to: str | None = None
    mention_level: LevelStr | None = None
    mention_if_ends: bool | None = None
    callsite_level: LevelStr | None = None
    callsite_context_before: int = 1
    callsite_context_after: int = 4
    verbose: bool | None = None
    disable: bool | None = None


DOC_ADDITIONS_BASE = {
    "send": lambda cls: f"""
        Example:

            .. code-block:: python

               # Immediately send "Job finished!" to your Slack channel
               {cls._platform.lower()}.send("Job finished!")

               # You can also send any Python data (it will be stringified)
               {cls._platform.lower()}.send(data)
        """,
    "watch": lambda cls: f"""
        Example:

            Monitor functions:

            .. code-block:: python

               # You can also optionally specify params to include in the notification
               # The values passed to these parameters are also reported
               @{cls._platform.lower()}.watch(params=["arg1", "arg2"])
               def long_task(arg1: int, arg2: str, arg3: bool) -> None:
                   # This function will be monitored
                   # You can receive notifications when it starts, ends, or encounters an error
                   ...
                   # Your long-running code here

            Monitor methods:

            .. code-block:: python

               with {cls._platform.lower()}.watch():
                   # Code inside this block will be monitored
                   # You can receive notifications when it starts, ends, or encounters an error
                   ...
                   # Your long-running code here

            Monitor Iterations (e.g., for loops):

            .. code-block:: python

                for i in {cls._platform.lower()}.watch(range(100), step=10):
                    # This loop will be monitored, and you'll receive notifications every 10 iterations.
                    ...
                    # Your long-running code here

        .. note::
           The above example does **not** catch exceptions automatically,
           since exceptions raised inside the for loop cannot be caught by the iterator in Python.
           If you also want to be notified when an error occurs, wrap your code in the monitoring context:

           .. code-block:: python

              with {cls._platform.lower()}.watch(range(100), step=10) as it:
                  for i in it:
                      # This loop will be monitored, and you'll receive notifications every 10 iterations.
                      # If an error occurs inside this context, you'll be notified immediately.
                      ...
                      # Your long-running code here
        """,
    "register": lambda cls: f"""
        Example:

            Monitor existing functions from libraries:

            .. code-block:: python

               import requests

               # Register the `get` function from the `requests` library
               {cls._platform.lower()}.register(requests, "get")

               # Now any time you call `requests.get`, it will be monitored
               response = requests.get("https://example.com/largefile.zip")

            Monitor existing methods of classes:

            .. code-block:: python

               from transformers import Trainer

               # Register the `train` method of the `Trainer` class
               {cls._platform.lower()}.register(Trainer, "train")

               # Now any time you call `trainer.train()`, it will be monitored
               trainer = Trainer(model=...)
               trainer.train()

            Monitor existing methods of specific class instances:

            .. code-block:: python

               from transformers import Trainer

               # Create a Trainer instance
               trainer = Trainer(model=...)

               # Register the `train` method of the `trainer` instance
               {cls._platform.lower()}.register(trainer, "train")

               # Now any time you call `trainer.train()`, it will be monitored
               trainer.train()
        """,
}


class BaseNotifier(ABC):
    """
    Abstract base class for all notifiers.

    Provides common functionality for sending messages and watching
    code execution, with optional exception handling and verbosity.
    """

    _platform: str
    """Name of the notification platform (e.g., "Slack")."""

    def __init__(
        self,
        channel: str | None = None,
        mention_to: str | None = None,
        mention_level: LevelStr = "error",
        mention_if_ends: bool = True,
        callsite_level: LevelStr = "error",
        token: str | None = None,
        verbose: bool | int = True,
        disable: bool = False,
    ) -> None:
        """
        Initialize the notifier with default settings.
        This settings can be overridden at each call of :meth:`register`, :meth:`send`, and :meth:`watch`.

        Args:
            channel:
                Default channel for notifications. If not provided, it will look for an environment variable
                named ``{platform}_CHANNEL`` where ``{platform}`` is the notifier's platform name in uppercase
                (e.g., ``SLACK_CHANNEL`` for Slack).
            mention_to:
                Default user to mention in notification. If not provided, it will look for an environment variable
                named ``{platform}_MENTION_TO`` where ``{platform}`` is the notifier's platform name in uppercase
                (e.g., ``SLACK_MENTION_TO`` for Slack).
            mention_level: Minimum log level to trigger a mention.
            mention_if_ends: Whether to mention at the end of the watch.
            callsite_level: Minimum log level to emit the call-site source snippet.
            token:
                API token or authentication key. If not provided, it will look for an environment variable named
                ``{platform}_BOT_TOKEN`` where ``{platform}`` is the notifier's platform name in uppercase
                (e.g., ``SLACK_BOT_TOKEN`` for Slack).
            verbose:
                If obj:`True`, log messages to console.
                If set to 1, only logs during initialization.
                If set to 2 or higher, behaves the same as obj:`True`.
            disable:
                If :obj:`True`, disable sending all notifications. This is useful for parallel runs or testing
                where you want to avoid sending actual notifications.

        .. note::
           The channel and token must be set, either via environment variables or as function arguments.
           If not set, the notification will not be sent, and an error will be logged
           (the original Python script will continue running without interruption).
        """
        self._mention_to = mention_to or os.getenv(
            f"{self._platform.upper()}_MENTION_TO"
        )
        self._token = token or os.getenv(f"{self._platform.upper()}_BOT_TOKEN")
        if not self._token and verbose:
            _log.error(
                f"Missing {self._platform} bot token. Please set the {self._platform.upper()}_BOT_TOKEN "
                "environment variable or pass it as an argument."
            )
            self._disable = True
        self._mention_level = mention_level
        self._mention_if_ends = mention_if_ends
        self._default_callsite_level = callsite_level
        self._default_channel = channel or os.getenv(
            f"{self._platform.upper()}_CHANNEL"
        )
        self._disable = disable
        if disable and verbose:
            _log.info(
                f"{self._platform}Notifier is disabled. No messages will be sent."
            )
        self._verbose = verbose if isinstance(verbose, bool) else verbose >= 2

    def send(
        self,
        data: Any,
        *,
        channel: str | None = None,
        mention_to: str | None = None,
        verbose: bool | None = None,
        disable: bool | None = None,
    ) -> None:
        """
        Send a notification message.
        You can send notifications at any point in your code, not just at the start or end of a task.
        Any data can be sent, and it will be stringified.

        Args:
            data: The payload or message content.
            channel: Override the default channel for notifications.
            mention_to: Override the default entity to mention on notification.
            verbose: Override the default verbosity setting.
            disable: Override the default disable flag.
        """
        self._send(
            _SendConfig(
                channel=channel or self._default_channel,
                mention_to=mention_to or self._mention_to,
                mention_level="info" if mention_to or self._mention_to else "error",
                verbose=verbose if verbose is not None else self._verbose,
                disable=disable if disable is not None else self._disable,
            ),
            str(data),
            prefix="Send message: ",
        )

    def _send(
        self,
        send_config: _SendConfig,
        message: str,
        tb: str | None = None,
        level: LevelStr = "info",
        prefix: str = "",
    ) -> None:
        try:
            if not send_config.disable:
                self._do_send(send_config, prepare_for_message(message), tb, level)
            if send_config.verbose:
                {
                    "info": _log.info,
                    "warning": _log.warn,
                    "error": _log.error,
                }[level](f"{prefix}{message}")
        except Exception as e:
            if send_config.verbose:
                _log.error(f"Error sending to {self._platform}: {e}")

    @abstractmethod
    def _do_send(
        self,
        send_config: _SendConfig,
        message: str,
        tb: str | None = None,
        level: LevelStr = "info",
    ) -> None:
        raise NotImplementedError

    @overload
    def watch(
        self,
        iterable: None = ...,
        /,
        *,
        params: str | list[str] | None = ...,
        step: int = ...,
        total: None = ...,
        **kwargs: Unpack[SendOptions],
    ) -> ContextManagerDecorator: ...

    @overload
    def watch(
        self,
        iterable: Iterable[T],
        /,
        *,
        params: None = ...,
        step: int = ...,
        total: int | None = ...,
        **kwargs: Unpack[SendOptions],
    ) -> ContextManagerIterator[T]: ...

    def watch(
        self,
        iterable: Iterable[T] | None = None,
        /,
        *,
        params: str | list[str] | None = None,
        step: int = 1,
        total: int | None = None,
        **options: Unpack[SendOptions],
    ) -> ContextManagerDecorator | ContextManagerIterator[T]:
        """
        If ``iterable`` is not provided, return an object that can serve as both a context manager and
        a decorator to watch code execution.
        This will automatically send notifications when the function or code block starts, ends,
        or raises an exception.

        If ``iterable`` is provided, return a generator that yields items from an ``iterable``
        while sending notifications about its progress.

        Args:
            iterable: An iterable (e.g., a list or range) to monitor progress.
            params:
                Names of the function parameters whose values should be included in the message
                when the decorated function is called.
                This option is ignored when used as a context manager.
            step:
                The number of items to process before sending a progress notification.
                This option is ignored if the iterable is not provided.
            total:
                The total number of items in the iterable.
                If not provided and the iterable has not ``__len__``,
                it will not be included in the progress messages.
                This option is ignored if the iterable is not provided.
            **options: Additional options. See :class:`~notist._notifiers.base.SendOptions` for details.

        Returns:
            An an object that can serve as both a context manager and a decorator.
        """
        return self._watch_impl(
            iterable,
            params=params,
            step=step,
            total=total,
            **options,
        )

    def _watch_impl(
        self,
        iterable: Iterable[T] | None = None,
        /,
        *,
        params: str | list[str] | None = None,
        step: int = 1,
        total: int | None = None,
        combined: int = 0,
        class_name: str | None = None,
        object_id: int | None = None,
        **options: Unpack[SendOptions],
    ) -> Watch | IterableWatch[T]:
        opts = _SendOptions(**options)
        send_config = _SendConfig(
            channel=opts.channel or self._default_channel,
            mention_to=opts.mention_to or self._mention_to,
            mention_level=opts.mention_level or self._mention_level,
            mention_if_ends=opts.mention_if_ends
            if opts.mention_if_ends is not None
            else self._mention_if_ends,
            verbose=opts.verbose if opts.verbose is not None else self._verbose,
            disable=opts.disable if opts.disable is not None else self._disable,
        )
        if iterable is None:
            return Watch(
                _SendFnPartial(self._send, send_config),
                params,
                opts.label,
                opts.callsite_level or self._default_callsite_level,
                opts.callsite_context_before,
                opts.callsite_context_after,
                combined,
            )
        else:
            if step < 1:
                step = 1
                if send_config.verbose:
                    _log.warn(
                        f"Step must be at least 1. Setting step to 1 for {self._platform}Notifier."
                    )

            return IterableWatch(
                iterable,
                step,
                total,
                _SendFnPartial(self._send, send_config),
                opts.label,
                opts.callsite_level or self._default_callsite_level,
                opts.callsite_context_before,
                opts.callsite_context_after,
                class_name,
                object_id,
            )

    def register(
        self,
        target: ModuleType | type[Any] | Any,
        name: str,
        params: str | list[str] | None = None,
        **options: Unpack[SendOptions],
    ) -> None:
        """
        Register existing function or method to be monitored by this notifier.
        This function corresponds to applying the :meth:`watch` decorator to an existing function or method.

        Args:
            target: The module, class, or class instance containing the function to be registered.
            name: The name of the function to be registered.
            params:
                Names of the function parameters whose values should be included in the message
                when the registered function is called.
            **options: Additional options. See :class:`~notist._notifiers.base.SendOptions` for details.
        """
        self._register_impl(target, name, params, **options)

    def _register_impl(
        self,
        target: ModuleType | type[Any] | Any,
        name: str,
        params: str | list[str] | None = None,
        combined: int = 0,
        **options: Unpack[SendOptions],
    ) -> None:
        opts = _SendOptions(**options)
        original = getattr(target, name, None)
        if original is None:
            if opts.verbose if opts.verbose is not None else self._verbose:
                _log.error(
                    f"Cannot register {self._platform}Notifier on `{target.__name__}.{name}`: "
                    f"target `{target.__name__}` has no attribute `{name}`."
                )
            return
        watch: Watch | IterableWatch = self._watch_impl(
            params=params,
            combined=combined,
            **options,
        )
        assert isinstance(watch, Watch)
        patched = watch(original)
        setattr(target, name, patched)
        target_name = (
            target.__name__
            if hasattr(target, "__name__")
            else f"<{target.__class__.__name__} object at {hex(id(target))}>"
        )
        if opts.verbose if opts.verbose is not None else self._verbose:
            _log.info(f"Registered {self._platform}Notifier on `{target_name}.{name}`.")



from __future__ import annotations

import re
import sys
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self


_PREFIX = "[notist] "


# NOTE: Python 3.12+ (PEP 695) supports type statement.
# After dropping Python 3.11 support, update this to use that instead.
# See:
#   - https://peps.python.org/pep-0695/
#   - https://docs.python.org/3/library/typing.html#type-aliases
LevelStr = Literal["info", "warning", "error"]
LEVEL_ORDER: dict[LevelStr, int] = {
    "info": 0,
    "warning": 1,
    "error": 2,
}


def info(message: str, with_timestamp: bool = True) -> None:
    _print_with_prefix(
        message,
        level_str="[INFO] ",
        prefix_color=fg256(48),
        time_color=fg256(14),
        with_timestamp=with_timestamp,
    )


def warn(message: str, with_timestamp: bool = True) -> None:
    _print_with_prefix(
        message,
        level_str="[WARN] ",
        prefix_color=fg256(214),
        time_color=fg16(93),
        with_timestamp=with_timestamp,
    )


def error(message: str, with_timestamp: bool = True) -> None:
    _print_with_prefix(
        message,
        level_str="[ERROR] ",
        prefix_color=fg256(196),
        time_color=fg256(213),
        with_timestamp=with_timestamp,
    )


def _print_with_prefix(
    message: str,
    level_str: str,
    prefix_color: str,
    time_color: str,
    with_timestamp: bool = True,
) -> None:
    prefix = f"{prefix_color}{_PREFIX}{level_str}{RESET}"
    prefix = (
        f"{prefix}{time_color}{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} {RESET}"
        if with_timestamp
        else prefix
    )
    message = "\n".join([prefix + line for line in message.splitlines()])
    print(_prepare_for_console(message))


_CSI = "\x1b["
RESET = f"{_CSI}0m"


def fg16(code: int) -> str:
    return f"{_CSI}{code}m"


def fg256(n: int) -> str:
    return f"{_CSI}38;5;{n}m"


# NOTE: Python 3.11+ introduces enum.StrEnum.
# After dropping Python 3.10 support, switch to stdlib StrEnum and remove the shim.
# See:
#   - https://docs.python.org/3/library/enum.html#enum.StrEnum
#   - https://docs.python.org/3.11/whatsnew/3.11.html
class Glyph(str, Enum):
    TL, TR, BL, BR = "╭", "╮", "╰", "╯"
    H, V, SEP_L, SEP_R, SEP_T, SEP_B = "─", "│", "├", "┤", "┬", "┴"
    BH = "━"
    TDH, BTDH, QDH, BQDH = "┄", "┅", "┈", "┉"
    RARROW, LARROW = "▶", "◀"
    RARROWF, LARROWF = "▷", "◁"
    RARROWP = "❯"
    BULLET, WBULLET, CBULLET = "•", "◦", "⦿"
    CHECK, CROSS = "✓", "✗"
    WARN, INFO = "⚠", "ℹ"

    def __str__(self) -> str:
        return self.value


@dataclass(frozen=True)
class _Expansion:
    placeholder: str
    console: str
    message: str


class SpecialToken(str, Enum):
    def __new__(cls, expansion: _Expansion) -> Self:
        obj = str.__new__(cls, expansion.placeholder)
        obj._value_ = expansion.placeholder
        obj._console = expansion.console  # type: ignore[attr-defined]
        obj._message = expansion.message  # type: ignore[attr-defined]
        return obj

    BT_CON = _Expansion("<|@BACKTICK::CONSOLE|>", "`", "")
    BT_MSG = _Expansion("<|@BACKTICK::MESSAGE|>", "", "`")
    BT_ALW = _Expansion("<|@BACKTICK::ALWAYS|>", "`", "`")
    FNO_CON = _Expansion("<|@FENCE!OPEN::CONSOLE|>", "```\n", "")
    FNC_CON = _Expansion("<|@FENCE!CLOSE::CONSOLE|>", "\n```", "")
    FNO_MSG = _Expansion("<|@FENCE!OPEN::MESSAGE|>", "", "```\n")
    FNC_MSG = _Expansion("<|@FENCE!CLOSE::MESSAGE|>", "", "\n```")
    FNO_ALW = _Expansion("<|@FENCE!OPEN::ALWAYS|>", "```\n", "```\n")
    FNC_ALW = _Expansion("<|@FENCE!CLOSE::ALWAYS|>", "\n```", "\n```")

    def __str__(self) -> str:
        return self.value


def _expand_special_tokens(text: str, target: Literal["console", "message"]) -> str:
    for token in SpecialToken:
        replacement = token._console if target == "console" else token._message  # type: ignore[attr-defined]
        text = text.replace(token.value, replacement)
    return text


def _strip_sgr(text: str) -> str:
    """Remove ANSI SGR sequences (ESC[ ... m)."""
    return re.compile(r"\x1b\[[0-9;]*m").sub("", text)


def _prepare_for_console(text: str) -> str:
    return _expand_special_tokens(text, target="console")


def prepare_for_message(text: str) -> str:
    text = _expand_special_tokens(text, target="message")
    return _strip_sgr(text)



from __future__ import annotations

from typing import TYPE_CHECKING

from slack_sdk import WebClient

from notist import _log
from notist._log import LEVEL_ORDER, RESET, fg256
from notist._notifiers.base import (
    DOC_ADDITIONS_BASE,
    BaseNotifier,
    _SendConfig,
)
from notist._utils import extend_method_docstring

if TYPE_CHECKING:
    from notist._log import LevelStr


_DOC_ADDITIONS = {
    "__init__": """
        Example:

            .. code-block:: python

               from notist import SlackNotifier

               # Create a SlackNotifier with defaults
               slack = SlackNotifier(
                   channel="my-channel",  # Slack channel name or ID
                   mention_to="@U012345678",  # Mention a specific user (Optional)
               )
        """,
}


@extend_method_docstring(_DOC_ADDITIONS | DOC_ADDITIONS_BASE)
class SlackNotifier(BaseNotifier):
    _platform = "Slack"

    def __init__(
        self,
        channel: str | None = None,
        mention_to: str | None = None,
        mention_level: LevelStr = "error",
        mention_if_ends: bool = True,
        callsite_level: LevelStr = "error",
        token: str | None = None,
        verbose: bool = True,
        disable: bool = False,
    ) -> None:
        super().__init__(
            channel,
            mention_to,
            mention_level,
            mention_if_ends,
            callsite_level,
            token,
            verbose,
            disable,
        )
        self._client = WebClient(token=self._token)
        if not self._disable and verbose:
            if self._default_channel:
                _log.info(
                    f"SlackNotifier initialized with default channel: {fg256(33)}{self._default_channel}{RESET}"
                )
            else:
                _log.warn(
                    "No Slack channel configured. Need to specify channel each time."
                )

    def _do_send(
        self,
        send_config: _SendConfig,
        message: str,
        tb: str | None = None,
        level: LevelStr = "info",
    ) -> None:
        channel = send_config.channel or self._default_channel
        if channel is None:
            if send_config.verbose:
                _log.error(
                    "No Slack channel specified.\nSkipping sending message to Slack."
                )
            return
        mention_to = send_config.mention_to or self._mention_to
        mention_level = send_config.mention_level or self._mention_level
        text = (
            f"<{mention_to}>\n{message}"
            if mention_to
            and (
                LEVEL_ORDER[level] >= LEVEL_ORDER[mention_level]
                or (send_config.mention_if_ends and "End" in message)
            )
            else message
        )
        self._client.chat_postMessage(
            text=text,
            channel=channel,
            attachments=tb
            and [
                {
                    "blocks": [
                        {
                            "type": "section",
                            "text": {
                                "type": "plain_text",
                                "text": tb,
                            },
                        }
                    ],
                    "color": "#ff3d33",
                }
            ],
        )



from __future__ import annotations

from contextlib import ContextDecorator
from typing import TYPE_CHECKING

import notist._core

from .notifiers_test.test_discord import dummy_post  # noqa: F401
from .notifiers_test.test_slack import dummy_client  # noqa: F401

if TYPE_CHECKING:
    import sys
    from typing import Any

    from pytest import CaptureFixture, MonkeyPatch

    from notist._notifiers.base import BaseNotifier

    from .notifiers_test.test_discord import Sent
    from .notifiers_test.test_slack import DummyClient

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

import pytest

from notist._core import (
    _DESTINATIONS,
    _allow_multi_dest,
    _combine_contexts,
)


@_allow_multi_dest
def dummy_send(
    send_to: _DESTINATIONS | list[_DESTINATIONS] | None = None, **kwargs: Any
) -> None:
    return None


@pytest.mark.parametrize(
    "send_to",
    [
        None,
        "slack",
        ["slack", "discord"],
    ],
)
def test_allow_multi_dest(send_to: _DESTINATIONS | list[_DESTINATIONS]) -> None:
    assert dummy_send(send_to=send_to) is None


def test_wo_specify_dest() -> None:
    assert dummy_send() is None


def test_combine_contexts_order() -> None:
    called = []

    class A(ContextDecorator):
        def __enter__(self) -> Self:
            called.append("A+")
            return self

        def __exit__(self, *e: Any) -> None:
            called.append("A-")

    class B(ContextDecorator):
        def __enter__(self) -> Self:
            called.append("B+")
            return self

        def __exit__(self, *e: Any) -> None:
            called.append("B-")

    combo = _combine_contexts([A(), B()])  # type: ignore
    with combo:
        called.append("BODY")
    assert called == ["A+", "B+", "BODY", "B-", "A-"]


parametrize_destination = pytest.mark.parametrize(
    "destination",
    ["slack", "discord", ["slack"], ["slack", "discord"]],
)


@parametrize_destination
def test_init(
    monkeypatch: MonkeyPatch,
    destination: _DESTINATIONS | list[_DESTINATIONS],
) -> None:
    dummy_notifiers: dict[str, BaseNotifier] = {}
    monkeypatch.setattr(notist._core, "_notifiers", dummy_notifiers)
    notist.init(send_to=destination, token="tok")
    if isinstance(destination, str):
        destination = [destination]
    assert set(dummy_notifiers.keys()) == set(destination)
    notist.init(send_to=destination, token="tok")
    assert set(dummy_notifiers.keys()) == set(destination)


@parametrize_destination
def test_send(
    monkeypatch: MonkeyPatch,
    capsys: CaptureFixture[str],
    dummy_client: DummyClient,  # noqa
    dummy_post: Sent,  # noqa
    destination: _DESTINATIONS | list[_DESTINATIONS],
) -> None:
    dummy_notifiers: dict[str, BaseNotifier] = {}
    monkeypatch.setattr(notist._core, "_notifiers", dummy_notifiers)
    notist.init(send_to=destination, token="tok", channel="chan")
    if isinstance(destination, str):
        destination = [destination]
    if "slack" in dummy_notifiers:
        dummy_notifiers["slack"]._client = dummy_client  # type: ignore
    notist.send("msg")
    if "slack" in destination:
        assert len(dummy_client.sent) == 1
        assert dummy_client.sent[0]["text"] == "msg"
    else:
        assert dummy_client.sent == []
    if "discord" in destination:
        assert len(dummy_post) == 1
        assert dummy_post[0][2]["content"] == "msg"
    else:
        assert len(dummy_post) == 0

    captured = capsys.readouterr()
    if "slack" == destination or "slack" in destination:
        assert "SlackNotifier initialized" in captured.out
    if "discord" == destination or "discord" in destination:
        assert "DiscordNotifier initialized" in captured.out
    assert captured.out.count("msg") == 1



from notist._core import init, register, send, watch
from notist._notifiers import (
    DiscordNotifier,
    SlackNotifier,
)

__all__ = [
    "DiscordNotifier",
    "SlackNotifier",
    "init",
    "register",
    "send",
    "watch",
]

__version__ = "0.1.0"



from __future__ import annotations

import contextlib
import inspect
import itertools
import sys
from collections.abc import Callable, Sequence
from contextlib import AbstractContextManager, ContextDecorator
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Literal,
    TypeVar,
    overload,
)

from notist import _log
from notist._notifiers.base import (
    BaseNotifier,
    ContextManagerDecorator,
    ContextManagerIterator,
    SendOptions,
)
from notist._notifiers.discord import DiscordNotifier
from notist._notifiers.slack import SlackNotifier

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

if TYPE_CHECKING:
    from collections.abc import Generator, Iterator
    from types import ModuleType, TracebackType

    from notist._log import LevelStr

    if sys.version_info >= (3, 10):
        from typing import TypeGuard
    else:
        from typing_extensions import TypeGuard

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    if sys.version_info >= (3, 12):
        from typing import Unpack
    else:
        from typing_extensions import Unpack


_notifiers: dict[str, BaseNotifier] = {}

_DESTINATIONS = Literal["slack", "discord"]
_DESTINATIONS_MAP: dict[_DESTINATIONS, type[BaseNotifier]] = {
    "slack": SlackNotifier,
    "discord": DiscordNotifier,
}


# NOTE: Python 3.12+ (PEP 695) supports inline type parameter syntax.
# After dropping Python 3.11 support, update this to use that instead.
# See:
#   - https://peps.python.org/pep-0695/
#   - https://docs.python.org/3/reference/compound_stmts.html#type-params
T = TypeVar("T")
_T = TypeVar("_T")
_P = ParamSpec("_P")
_F = TypeVar("_F", bound=Callable[..., Any])

# # NOTE: Python 3.10+ (PEP 604) supports writing union types with `X | Y`.
# # After dropping Python 3.9 support, we can remove using `typing.Union`.
# # See:
# #   - https://peps.python.org/pep-0604/
# #   - https://docs.python.org/3/library/stdtypes.html#types-union
if sys.version_info >= (3, 10):
    _R = ContextManagerDecorator | ContextManagerIterator | None
else:
    from typing import Union

    _R = Union[ContextManagerDecorator, ContextManagerIterator, None]


@overload
def _allow_multi_dest(
    fn: Callable[_P, ContextManagerDecorator | ContextManagerIterator],
) -> Callable[_P, ContextManagerDecorator | ContextManagerIterator]: ...
@overload
def _allow_multi_dest(fn: Callable[_P, None]) -> Callable[_P, None]: ...


def _allow_multi_dest(fn: Callable[_P, _R]) -> Callable[_P, _R]:
    @wraps(fn)
    def wrapper(
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _R:
        send_to = kwargs.get("send_to")
        if send_to is None and _notifiers:
            send_to = list(_notifiers.keys())
        iterable = kwargs.get("iterable")
        if isinstance(send_to, Sequence) and not isinstance(send_to, str):
            res = []
            for i, dest in enumerate(send_to):
                new_kwargs = kwargs.copy()
                new_kwargs["send_to"] = dest
                if i and iterable is not None:
                    new_kwargs["iterable"] = itertools.repeat(None)
                    new_kwargs["class_name"] = iterable.__class__.__name__
                    new_kwargs["object_id"] = hex(id(iterable))
                if i:
                    new_kwargs["verbose"] = 1
                if "combined" in inspect.signature(fn).parameters.keys():
                    new_kwargs["combined"] = len(send_to) - i
                res.append(fn(*args, **new_kwargs))  # type: ignore
            if _are_all_contexts(res):
                return _combine_contexts(res)
            elif all(r is None for r in res):
                return None
            else:
                raise ValueError("Cannot mix.")
        else:
            return fn(*args, **kwargs)

    return wrapper


def _combine_contexts(
    contexts: Sequence[AbstractContextManager],
) -> ContextManagerDecorator | ContextManagerIterator:
    class _CombinedBase(AbstractContextManager):
        def __enter__(self) -> Self:
            for ctx in contexts:
                ctx.__enter__()
            return self

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            tb: TracebackType | None,
        ) -> None:
            for ctx in reversed(contexts):
                ctx.__exit__(exc_type, exc_val, tb)

    combined_cls: type[_CombinedBase]
    # cannot use TypeGuard here
    if all(callable(ctx) for ctx in contexts):

        class _CombinedContextManagerDecorator(_CombinedBase):
            def __call__(self, fn: _F) -> _F:
                wrapped = fn
                for ctx in contexts:
                    assert callable(ctx)
                    wrapped = ctx(wrapped)
                return wrapped

        combined_cls = _CombinedContextManagerDecorator
    elif all(hasattr(ctx, "__iter__") for ctx in contexts):

        class _CombinedContextManagerIterator(_CombinedBase):
            def __iter__(self) -> Iterator:
                return map(lambda x: x[0], zip(*contexts))

        combined_cls = _CombinedContextManagerIterator
    else:
        raise ValueError("Cannot mix context decorators and context iterators.")

    return combined_cls()


def _are_all_contexts(
    contexts: Sequence[object],
) -> TypeGuard[Sequence[AbstractContextManager]]:
    return all(isinstance(ctx, AbstractContextManager) for ctx in contexts)


class _PhantomContextManagerDecorator(ContextDecorator, contextlib.nullcontext):
    """A no-op context manager decorator that does nothing."""


class _PhantomContextManagerIterator(Iterable[_T], contextlib.nullcontext):
    """A no-op context manager iterator that does nothing."""

    def __init__(self, iterable: Iterable[_T]) -> None:
        self._iterable = iterable

    def __iter__(self) -> Generator[_T, None, None]:
        yield from self._iterable


@_allow_multi_dest
def init(
    *,
    send_to: _DESTINATIONS | list[_DESTINATIONS],
    channel: str | None = None,
    mention_to: str | None = None,
    mention_level: LevelStr = "error",
    mention_if_ends: bool = True,
    callsite_level: LevelStr = "error",
    token: str | None = None,
    verbose: bool | int = True,
    disable: bool = False,
) -> None:
    """
    Initialize the notifier with default settings.
    This settings can be overridden at each call of :func:`~notist._core.register`,
    :func:`~notist._core.send`, and :func:`~notist._core.watch`.
    Alternatively, you can skip initialization with this function and provide all settings directly through
    these functions.

    Args:
        send_to: Destination(s) to send notifications to. e.g., "slack", "discord", or ["slack", "discord"].
        channel:
            Default channel for notifications. If not provided, it will look for an environment variable
            named ``{platform}_CHANNEL`` where ``{platform}`` is the notifier's platform name in uppercase
            (e.g., ``SLACK_CHANNEL`` for Slack).
        mention_to:
            Default user to mention in notification. If not provided, it will look for an environment variable
            named ``{platform}_MENTION_TO`` where ``{platform}`` is the notifier's platform name in uppercase
            (e.g., ``SLACK_MENTION_TO`` for Slack).
        mention_level: Minimum log level to trigger a mention.
        mention_if_ends: Whether to mention at the end of the watch.
        callsite_level: Minimum log level to emit the call-site source snippet.
        token:
            API token or authentication key. If not provided, it will look for an environment variable named
            ``{platform}_BOT_TOKEN`` where ``{platform}`` is the notifier's platform name in uppercase
            (e.g., ``SLACK_BOT_TOKEN`` for Slack).
        verbose:
            If obj:`True`, log messages to console.
            If set to 1, only logs during initialization.
            If set to 2 or higher, behaves the same as obj:`True`.
        disable:
            If :obj:`True`, disable sending all notifications. This is useful for parallel runs or testing
            where you want to avoid sending actual notifications.

    .. note::
       The channel and token must be set, either via environment variables or as function arguments.
       If not set, the notification will not be sent, and an error will be logged
       (the original Python script will continue running without interruption).

    .. note::
       The destination (``send_to``) must be set, either in this :func:`~notist._core.init` function
       or as an argument to subsequent calls.

    Example:

        .. code-block:: python

           import notist

           # Set up Slack notifiers with defaults
           notist.init(send_to="slack", channel="my-channel", mention_to="@U012345678")
    """
    global _notifiers
    assert isinstance(send_to, str)
    if send_to in _notifiers:
        _log.warn(
            f"{_DESTINATIONS_MAP[send_to].__name__} already initialized. Skipping."
        )
        return
    _notifiers[send_to] = _DESTINATIONS_MAP[send_to](
        channel=channel,
        mention_to=mention_to,
        mention_level=mention_level,
        mention_if_ends=mention_if_ends,
        callsite_level=callsite_level,
        token=token,
        verbose=verbose,
        disable=disable,
    )


@_allow_multi_dest
def send(
    data: Any,
    *,
    send_to: _DESTINATIONS | list[_DESTINATIONS] | None = None,
    channel: str | None = None,
    mention_to: str | None = None,
    verbose: bool | None = None,
    disable: bool | None = None,
) -> None:
    """
    Send a notification message.
    You can send notifications at any point in your code, not just at the start or end of a task.
    Any data can be sent, and it will be stringified.

    Args:
        data: The payload or message content.
        send_to: Destination(s) to send notifications to. e.g., "slack", "discord", or ["slack", "discord"].
        channel: Override the default channel for notifications.
        mention_to: Override the default entity to mention on notification.
        verbose: Override the default verbosity setting.
        disable: Override the default disable flag.

    Example:

        .. code-block:: python

           # Immediately send "Job finished!" to your Slack channel
           notist.send("Job finished!")

           # You can also send any Python data (it will be stringified)
           notist.send(data)
    """
    if send_to is None:
        _warn_not_set_send_to()
        return
    assert isinstance(send_to, str)
    init_opts = SendOptions(
        channel=channel,
        mention_to=mention_to,
        verbose=verbose,
        disable=disable,
    )
    _init_if_needed(send_to, init_opts)
    _notifiers[send_to].send(data, **init_opts)  # type: ignore


@overload
def watch(
    iterable: None = ...,
    /,
    *,
    send_to: _DESTINATIONS | list[_DESTINATIONS] | None = ...,
    step: int = ...,
    total: None = ...,
    params: str | list[str] | None = ...,
    **options: Unpack[SendOptions],
) -> ContextManagerDecorator: ...


@overload
def watch(
    iterable: Iterable[T],
    /,
    *,
    send_to: _DESTINATIONS | list[_DESTINATIONS] | None = ...,
    step: int = ...,
    total: int | None = ...,
    params: None = ...,
    **options: Unpack[SendOptions],
) -> ContextManagerIterator[T]: ...


def watch(
    iterable: Iterable[T] | None = None,
    /,
    *,
    send_to: _DESTINATIONS | list[_DESTINATIONS] | None = None,
    params: str | list[str] | None = None,
    step: int = 1,
    total: int | None = None,
    **options: Unpack[SendOptions],
) -> ContextManagerDecorator | ContextManagerIterator[T]:
    """
    Return an object that can serve as both a context manager and a decorator to watch code execution.
    This will automatically send notifications when the function or code block starts, ends, or raises
    an exception.

    Args:
        iterable: An iterable (e.g., a list or range) to monitor progress.
        params:
            Names of the function parameters whose values should be included in the message
            when the decorated function is called.
            This option is ignored when used as a context manager.
        label:
            Optional label for the watch context.
            This label will be included in both notification messages and log entries.
        send_to: Destination(s) to send notifications to. e.g., "slack", "discord", or ["slack", "discord"].
        **options: Additional options. See :class:`~notist._notifiers.base.SendOptions` for details.

    Returns:
        An an object that can serve as both a context manager and a decorator.

    Example:

        Use as a decorator to monitor a function:

        .. code-block:: python

           @notist.watch()
           def long_task():
               # This function will be monitored
               # Your long-running code here
               ...

        Use as a context manager to monitor a block of code:

        .. code-block:: python

           with notist.watch():
               # Code inside this block will be monitored
               # Your long-running code here
               ...

        Use to monitor an iterable in a for loop:

        .. code-block:: python

           # Monitor progress of processing a long-running for loop
           for i in notist.watch(range(100), step=10):
               # This loop will be monitored, and you'll receive notifications every 10 iterations.
               # If an error occurs inside this loop, you'll be notified immediately.
               ...

    .. note::
       The above example does **not** catch exceptions automatically,
       since exceptions raised inside the for loop cannot be caught by the iterator in Python.
       If you also want to be notified when an error occurs, wrap your code in the monitoring context:

       .. code-block:: python

           with notist.watch(range(100), step=10) as it:
               for i in it:
                   # This loop will be monitored, and you'll receive notifications every 10 iterations.
                   # If an error occurs inside this context, you'll be notified immediately.
                   ...
    """
    return _watch_impl(
        iterable,
        send_to=send_to,
        params=params,
        step=step,
        total=total,
        **options,
    )


@_allow_multi_dest
def _watch_impl(
    iterable: Iterable[T] | None = None,
    /,
    *,
    send_to: _DESTINATIONS | list[_DESTINATIONS] | None = None,
    params: str | list[str] | None = None,
    step: int = 1,
    total: int | None = None,
    combined: int = 0,
    class_name: str | None = None,
    object_id: int | None = None,
    **options: Unpack[SendOptions],
) -> ContextManagerDecorator | ContextManagerIterator[T]:
    if send_to is None:
        _warn_not_set_send_to()
        return (
            _PhantomContextManagerDecorator()
            if iterable is None
            else _PhantomContextManagerIterator(iterable)
        )
    assert isinstance(send_to, str)
    _init_if_needed(send_to, options)
    return _notifiers[send_to]._watch_impl(
        iterable,
        params=params,
        step=step,
        total=total,
        combined=combined,
        class_name=class_name,
        object_id=object_id,
        **options,
    )


def register(
    target: ModuleType | type[Any] | Any,
    name: str,
    params: str | list[str] | None = None,
    *,
    send_to: _DESTINATIONS | list[_DESTINATIONS] | None = None,
    **options: Unpack[SendOptions],
) -> None:
    """
    Register existing function or method to be watched by this notifier.
    This function corresponds to applying the :meth:`watch` decorator to an existing function or method.

    Args:
        target: The module, class, or class instance containing the function to be registered.
        name: The name of the function to be registered.
        params:
            Names of the function parameters whose values should be included in the message
            when the registered function is called.
        send_to:
            Destination(s) to send notifications to. e.g., "slack", "discord", or ["slack", "discord"].
        **options: Additional options. See :class:`~notist._notifiers.base.SendOptions` for details.

    Example:

        Monitor existing functions from libraries:

        .. code-block:: python

           import requests

           # Register the `get` function from the `requests` library
           notist.register(requests, "get")

           # Now any time you call `requests.get`, it will be monitored
           response = requests.get("https://example.com/largefile.zip")

        Monitor existing methods of classes:

        .. code-block:: python

           from transformers import Trainer

           # Register the `train` method of the `Trainer` class
           notist.register(Trainer, "train")

           # Now any time you call `trainer.train()`, it will be monitored
           trainer = Trainer(model=...)
           trainer.train()

        Monitor existing methods of specific class instances:

        .. code-block:: python

           from transformers import Trainer

           # Create a Trainer instance
           trainer = Trainer(model=...)

           # Register the `train` method of the `trainer` instance
           notist.register(trainer, "train")

           # Now any time you call `trainer.train()`, it will be monitored
           trainer.train()
    """
    _register_impl(
        target,
        name,
        params,
        send_to=send_to,
        **options,
    )


@_allow_multi_dest
def _register_impl(
    target: ModuleType | type[Any] | Any,
    name: str,
    params: str | list[str] | None = None,
    *,
    send_to: _DESTINATIONS | list[_DESTINATIONS] | None = None,
    combined: int = 0,
    **options: Unpack[SendOptions],
) -> None:
    if send_to is None:
        _warn_not_set_send_to()
        return None
    assert isinstance(send_to, str)
    _init_if_needed(send_to, options)
    _notifiers[send_to]._register_impl(
        target,
        name,
        params,
        combined=combined,
        **options,
    )


def _init_if_needed(
    send_to: _DESTINATIONS | list[_DESTINATIONS],
    opts: SendOptions,
) -> None:
    if send_to not in _notifiers:
        init(
            send_to=send_to,
            **{  # type: ignore
                k: v
                for k, v in opts.items()
                if v is not None and v in inspect.signature(init).parameters.keys()
            },
        )
    _update_verbose(opts)


def _update_verbose(opts: SendOptions) -> None:
    opts["verbose"] = (verbose := opts.get("verbose")) and (
        verbose if isinstance(verbose, bool) else verbose >= 2
    )


def _warn_not_set_send_to() -> None:
    _log.warn(
        "No destination specified. "
        "Please specify `send_to` parameter or initialize notifier with `notist.init()`. "
        "No notifications will be sent."
    )
